@@ -16,45 +16,46 @@ import functools
 import time
 import traceback
 import uuid
+from collections import defaultdict
 from contextlib import ContextDecorator
 from dataclasses import dataclass, field
-from typing import Any, Callable, Dict, Optional, Union
+from logging import Filter, Logger
+from threading import get_ident
+from typing import Any, Callable, Dict, Optional, TypeVar, Union
 
 from ondewo.logging.constants import CONTEXT, EXCEPTION, FINISH, START
+from ondewo.logging.filters import ThreadContextFilter
 from ondewo.logging.logger import logger_console
 
+TF = TypeVar("TF", bound=Callable[..., Any])
+
 
 @dataclass
 class Timer(ContextDecorator):
     """Time your code using a class, context manager, or decorator"""
 
-    name: Optional[str] = None
+    name: str = field(default_factory=lambda: str(uuid.uuid4()))
     message: str = FINISH
-    logger: Optional[
-        Callable[[Union[str, Dict[str, Any]]], None]
-    ] = logger_console.warning
-    _start_time: Optional[float] = field(default=None, init=False, repr=False)
+    logger: Callable[..., None] = logger_console.warning
+    _start_times: Dict[int, float] = field(default_factory=dict, init=False, repr=False)
     log_arguments: bool = True
     suppress_exceptions: bool = False
     recursive: bool = False
-    recurse_depth: int = 0
+    recurse_depths: Dict[int, float] = field(default_factory=lambda: defaultdict(float))
     argument_max_length: int = 10000
 
-    def __post_init__(self) -> None:
-        """Initialization: add timer to dicCallable[[Union[str, Dict[str, Any]]], None]t of timers"""
-        if not self.name:
-            self.name = str(uuid.uuid4())
+    def __call__(self, func: TF) -> TF:
+        """Decorator which adds a logs timing information for the decorated function.
 
-    def __call__(self, func: Callable) -> Callable:  # type: ignore
-        """
-        Decorator which adds a logs timing information for the decorated function.
+        Args
+            func: the function to be decorated
 
-        :param func:    the function to be decorated
-        :return:        the decorator
+        Returns:
+            the decorator
         """
 
         @functools.wraps(func)
-        def wrapper_timing(*args, **kwargs) -> Any:  # type: ignore
+        def wrapper_timing(*args, **kwargs) -> Any:
             self.start(func)
 
             try:
@@ -75,42 +76,57 @@ class Timer(ContextDecorator):
             self.stop(func.__name__)
             return value
 
-        return wrapper_timing
+        return wrapper_timing  # type: ignore
 
-    def start(self, func: Optional[Callable] = None) -> None:
+    def start(self, func: Optional[TF] = None) -> None:
         """Start a new timer"""
+        thread_id: int = get_ident()
         if func:
-            self.logger({"message": START.format(func.__name__)})  # type: ignore
-        if self._start_time is not None:
+            self.logger({"message": START.format(func.__name__, thread_id)})
+        if thread_id in self._start_times:
             if self.recursive:
-                self.recurse_depth += 1
-                self.logger(f"Recursing, depth = {self.recurse_depth}")  # type: ignore
+                self.recurse_depths[thread_id] += 1
+                self.logger(f"Recursing, depth = {self.recurse_depths[thread_id]}")
                 return
         else:
-            self._start_time = time.perf_counter()
+            self._start_times[thread_id] = time.perf_counter()
 
     def stop(self, func_name: Optional[str] = None) -> float:
         """Stop the timer, and report the elapsed time"""
-        if self.recurse_depth:
-            self.recurse_depth -= 1
+        thread_id: int = get_ident()
+
+        if self.recurse_depths[thread_id]:
+            self.recurse_depths[thread_id] -= 1
             if self.recursive:
                 return 0.0
 
         # Calculate elapsed time
-        assert self._start_time
-        elapsed_time = time.perf_counter() - self._start_time
+        if thread_id in self._start_times:
+            elapsed_time = time.perf_counter() - self._start_times[thread_id]
+
+            # reset the start time
+            del self._start_times[thread_id]
+        else:
+            elapsed_time = 0.0
 
         # Report elapsed time
         if self.logger:
-            self.report(elapsed_time, func_name)
+            self.report(
+                elapsed_time=elapsed_time, func_name=func_name, thread_id=thread_id
+            )
 
         return elapsed_time
 
-    def report(self, elapsed_time: float, func_name: Optional[str]) -> None:
-        name = func_name if func_name else CONTEXT
+    def report(
+        self,
+        elapsed_time: float,
+        func_name: Optional[str] = None,
+        thread_id: Optional[int] = None,
+    ) -> None:
+        name: str = func_name or CONTEXT
 
         try:
-            message = self.message.format(elapsed_time, name)
+            message = self.message.format(elapsed_time, name, thread_id)
         except IndexError:
             pass
 
@@ -222,12 +238,10 @@ def log_exception(
 
 
 def log_args_kwargs_results(  # type: ignore
-    func: Callable,
+    func: TF,
     result: Any,
     argument_max_length: int = -1,
-    logger: Optional[
-        Callable[[Union[str, Dict[str, Any]]], None]
-    ] = logger_console.warning,
+    logger: Optional[Callable[..., None]] = logger_console.warning,
     *args,
     **kwargs,
 ) -> None:
@@ -258,3 +272,29 @@ def log_args_kwargs_results(  # type: ignore
                 **formatted_results,
             }
         )
+
+
+class ThreadContextLogger(ContextDecorator):
+    """Add per-thread context information using a class, context manager or decorator."""
+
+    def __init__(
+        self,
+        context_dict: Optional[Dict[str, Any]] = None,
+        logger: Optional[Logger] = None,
+    ) -> None:
+        """
+
+        Args:
+            context_dict: optional context information to add to the logs from the current thread
+            logger: optional logger to add the information to (be default the global logger_console)
+        """
+        self.logger: Logger = logger or logger_console
+        self.filter: Filter = ThreadContextFilter(context_dict=context_dict)
+
+    def __enter__(self) -> None:
+        """Add the filter to the logger when entering the context."""
+        self.logger.addFilter(self.filter)
+
+    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
+        """Remove the filter from the logger when leaving the context."""
+        self.logger.removeFilter(self.filter)
